--- a/catenate.pl	2013-07-23 08:15:45.000000000 -0500
+++ b/catenate.pl	2013-08-05 13:29:08.000000000 -0500
@@ -575,9 +575,10 @@
 		failif($parts == 0, "0 parts");
 		send_data("CATENATE (");
 		$$size_ref = 0;
+		my $deferred_fail = undef;
 		for (1..$parts) {
 			failif($_ > $MAX_CATENATE_PARTS, "too many parts");
-			my $ret = part($tag, $size_ref);
+			my $ret = part($tag, $size_ref, \$deferred_fail);
 			return $ret if $ret <= 0;
 			send_data(" ") if $_ < $parts;
 		}
@@ -587,6 +588,8 @@
 		# dovecot >= 2.2 (which doesn't) doesn't
 		failif($imap_auth_x_plain_submit && $$size_ref == 0, "empty message");
 
+		failif(defined $deferred_fail, $deferred_fail);
+
 		# 0 parts is invalid syntax, so don't MULTIAPPEND after it
 		return 2 if $parts == 0;
 	}
@@ -598,6 +601,7 @@
 {
 	my $tag = shift;
 	my $size_ref = shift;
+	my $deferred_fail_ref = shift;
 
 	if (int(rand(2)) == 0) {
 		my @good_urls = keys %template_urls;
@@ -636,7 +640,15 @@
 			return $ret if $ret <= 0;
 		}
 
-		failif($bad, "bad url");
+		if ($bad) {
+			if ($imap_auth_x_plain_submit) {
+				# dovecot < 2.2 (which offers X-PLAIN-SUBMIT auth) rejects immediately after the first bad URL
+				failif(1, "bad url");
+			} else {
+				# dovecot >= 2.2 (which doesn't) rejects after it finishes reading the last contiguous url
+				$$deferred_fail_ref = "bad url";
+			}
+		}
 		failif($$size_ref > $MAX_CATENATE_MSG_SIZE, "message too large");
 	} else {
 		my @texts = ($smallbody, $largebody, "");
@@ -645,6 +657,8 @@
 			$text = "X-Catenate-Append-Tag: $tag\r\n$text";
 		}
 
+		failif(defined $$deferred_fail_ref, $$deferred_fail_ref);
+
 		$$size_ref += length($text);
 		failif($$size_ref > $MAX_CATENATE_MSG_SIZE, "message too large");
 
